import hashlib
from typing import Any, Generator, Type, Union

from tarn.compat import HashAlgorithm
from tarn.pickler import dumps

from ..containers import EdgesBag
from ..engine import (
    Command, Details, Node, NodeHash, NodeHashes, Request, Response, StaticGraph, StaticHash, CustomHash, LeafHash
)
from ..utils import StringsLike
from .base import CallableLayer
from .cache import to_seq


class HashDigest(CallableLayer):
    def __init__(self, names: StringsLike, algorithm: Union[Type[HashAlgorithm], str, None] = None,
                 return_value: bool = False):
        if isinstance(algorithm, str):
            algorithm = getattr(hashlib, algorithm)

        names = to_seq(names)
        details = Details(type(self))
        inputs, outputs, edges = [], [], []
        for name in names:
            inp, out = Node(name, details), Node(name, details)
            inputs.append(inp)
            outputs.append(out)
            edges.append(HashDigestEdge(algorithm, return_value).bind(inp, out))

        super().__init__(EdgesBag(
            inputs, outputs, edges,
            context=None, virtual=None, persistent=None, optional=None,
        ), ())


class HashDigestEdge(StaticGraph, StaticHash):
    def __init__(self, algorithm, return_value):
        super().__init__(arity=1)
        self.algorithm = algorithm
        self.return_value = return_value

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.HashDigest', LeafHash(self.algorithm), LeafHash(self.return_value), *inputs)

    def evaluate(self) -> Generator[Request, Response, Any]:
        result = []
        if self.return_value:
            value = yield Command.ParentValue, 0
            result.append(value)

        node_hash = yield Command.ParentHash, 0
        result.append(node_hash)
        node_hash = node_hash.value

        pickled = dumps(node_hash)
        result.append(pickled)
        if self.algorithm is not None:
            result.append(self.algorithm(pickled).digest())

        return tuple(result)



from typing import Any, Sequence

from ..containers import EdgesBag
from ..engine import Details, Node, StaticEdge, StaticGraph
from ..engine.node_hash import NodeHash, NodeHashes
from ..utils import node_to_dict
from .base import EdgesBag, Layer


class CheckIds(Layer):
    """
    Raise FieldError if id is not in ids
    """
    def __repr__(self):
        return 'CheckIds()'

    def _connect(self, previous: EdgesBag) -> EdgesBag:
        copy = previous.freeze()
        details = Details(type(self))
        edges, inputs, outputs = list(copy.edges), copy.inputs, node_to_dict(copy.outputs)
        assert len(inputs) == 1, 'Only one input is allowed'
        new_input = Node(inputs[0].name, details)
        edges.append(CheckIdsEdge().bind([new_input, outputs['ids']], inputs[0]))
        return EdgesBag([new_input], outputs.values(), edges, copy.context,
                        persistent=copy.persistent, optional=copy.optional,
                        virtual=copy.virtual)


class CheckIdsEdge(StaticGraph, StaticEdge):
    def __init__(self):
        super().__init__(arity=2)

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return inputs[0]

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        id_, ids = inputs
        if id_ in ids:
            return id_
        raise KeyError(f'{id_} is not in ids')



from collections import defaultdict

from typing import Callable, Union, Sequence, Any


from ..cache import MemoryCache
from ..containers import EdgesBag
from ..engine import (
    Details, Node, FunctionEdge, StaticGraph, StaticHash, Graph, NodeHashes, NodeHash, CustomHash, StaticEdge, CacheEdge
)
from ..engine import EvalGen, Command
from ..utils import node_to_dict, extract_signature
from .chain import connect
from .dynamic import DynamicConnectLayer
from .join import to_hash_id


class GroupBy(DynamicConnectLayer):
    def __init__(self, by: Union[str, Sequence[str], Callable]):
        self.by = by

    def _by_layer(self, details):
        output = Node('$by', details)
        if isinstance(self.by, str):
            inputs = [Node(self.by, details)]
            edges = [FunctionEdge(to_key, len(inputs)).bind(inputs, output)]
        elif callable(self.by):
            inputs, _ = extract_signature(self.by)
            inputs = [Node(x, details) for x in inputs]
            func = Node('$func', details)
            edges = [
                FunctionEdge(self.by, len(inputs)).bind(inputs, func),
                FunctionEdge(to_key, 1).bind(func, output),
            ]
        else:
            inputs = [Node(x, details) for x in self.by]
            edges = [FunctionEdge(to_key, len(inputs)).bind(inputs, output)]

        return EdgesBag(inputs, [output], edges, None)

    def _prepare_container(self, previous: EdgesBag) -> EdgesBag:
        details = Details(type(self))
        main = previous.freeze(details)

        assert len(main.inputs) == 1, main.inputs
        edges = list(main.edges)
        outputs = []
        key_name, keys_name = 'id', 'ids'
        changed_input = Node(key_name, details)
        mapping_node = Node('$mapping', details)
        output_nodes = node_to_dict(main.outputs)
        assert keys_name in output_nodes
        keys = output_nodes[keys_name]
        outputs.append(changed_input)

        # create a mapping: {new_id: [old_ids]} and store it in memory
        by_layer = self._by_layer(details)
        raw_mapping = Node('$mapping', details)
        edges.append(GroupMapping(
            connect(main, by_layer).compile().compile('$by')
        ).bind(keys, raw_mapping))
        edges.append(CacheEdge(MemoryCache(None)).bind(raw_mapping, mapping_node))

        compiler = main.compile()
        # evaluate each output
        for node in main.outputs:
            if node.name in [keys_name, key_name]:
                continue

            output = Node(node.name, details)
            outputs.append(output)
            edges.append(GroupEdge(compiler.compile(node.name)).bind([changed_input, mapping_node], output))

        if len(outputs) == 1:
            raise RuntimeError('Previous layer must contain at least 2 fields in order to perform a GroupBy operation')

        # update ids
        output_ids = Node(keys_name, details)
        outputs.append(output_ids)
        edges.append(FunctionEdge(lambda x: tuple(sorted(x)), arity=1).bind(mapping_node, output_ids))

        return EdgesBag(
            [changed_input], outputs, edges, None, persistent=main.persistent,
            optional=main.optional, virtual=None,
        )


class GroupEdge(StaticGraph, StaticEdge):
    def __init__(self, graph):
        super().__init__(arity=2)
        self.graph = graph
        self._hash = self.graph.hash()

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.GroupEdge', self._hash, *inputs)

    def _evaluate(self, inputs: Sequence[Any]) -> Any:
        """ arguments: id, mapping """
        # get the required ids
        new_key, mapping = inputs
        if new_key not in mapping:
            raise KeyError(f'The key {new_key} is not found')
        return {old_key: self.graph(old_key) for old_key in sorted(mapping[new_key])}


class GroupMapping(StaticGraph, StaticHash):
    def __init__(self, graph: Graph):
        super().__init__(arity=1)
        self.graph = graph
        self._hash = graph.hash()

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.GroupMapping', self._hash, *inputs)

    def evaluate(self) -> EvalGen:
        keys = yield Command.ParentValue, 0
        mapping = defaultdict(set)
        for key in keys:
            new = self.graph(key)
            assert key not in mapping[new], (key, mapping[new])
            mapping[new].add(key)

        return dict(mapping)


def to_key(*args) -> str:
    assert args
    if len(args) > 1:
        return to_hash_id(list(map(to_key, args)))
    x, = args
    if isinstance(x, str):
        return x
    if isinstance(x, (list, tuple)):
        return to_key(*x)

    raise TypeError(f"Can't convert an object of type {type(x).__name__!r} to key")



import logging
from typing import Callable, Collection, Dict, Iterable, Tuple, Type, Union

from ..layers import CallableLayer, Layer
from ..utils import MultiDict
from .factory import GraphFactory, SourceFactory, TransformFactory, add_from_mixins, add_quals, items_to_container

logger = logging.getLogger(__name__)
BASES: Dict[Type[Layer], GraphFactory] = {}


class APIMeta(type):
    @classmethod
    def __prepare__(mcs, *args, **kwargs):
        return MultiDict()

    def __new__(mcs, class_name, bases, namespace, **flags):
        if '__factory' in flags:
            factory = flags.pop('__factory')
            scope = namespace.to_dict()
            base = super().__new__(mcs, class_name, bases, scope, **flags)
            BASES[base] = factory  # noqa
            return base

        bases = set(bases)
        intersection = set(BASES) & bases
        if len(intersection) != 1:
            raise TypeError(f'Layers must inherit from one of ' + ', '.join(x.__name__ for x in BASES))

        main, = intersection
        factory = BASES[main]
        bases -= intersection
        base_name = main.__name__
        for base in bases:
            if not issubclass(base, Mixin):
                raise TypeError(f'{base_name}s can only inherit directly from "{base_name}" or other mixins.')

        logger.info('Compiling the layer "%s" of type %s', class_name, base_name)

        if main == Mixin:
            add_from_mixins(namespace, bases)
            scope = add_quals({'__methods__': namespace}, namespace)

        else:
            factory.validate_before_mixins(namespace)
            add_from_mixins(namespace, bases)
            scope = factory.make_scope(class_name, namespace)

        return super().__new__(mcs, class_name, (main,), scope, **flags)


class Source(CallableLayer, metaclass=APIMeta, __factory=SourceFactory):
    """
    Base class for all sources.
    """

    def __init__(self, *args, **kwargs):  # noqa
        raise RuntimeError("\"Source\" can't be directly initialized. You must subclass it first.")


class SourceBase(CallableLayer):
    def __init__(self, items: Union[Iterable[Tuple[str, Callable]], Dict[str, Callable]]):
        super().__init__(*items_to_container(items, type(self), SourceFactory))


class Transform(CallableLayer, metaclass=APIMeta, __factory=TransformFactory):
    """
    Base class for all transforms.

    Can also be used as an inplace factory for transforms.

    Examples
    --------
    # class-based transforms
    >>> from imops import zoom
    >>>
    >>> class Zoom(Transform):
    ...     def image(image):
    ...         return zoom(image, scale_factor=2)
    # inplace transforms
    >>> Transform(image=lambda image: zoom(image, scale_factor=2))
    """
    __inherit__: Union[str, Collection[str], bool] = ()
    __exclude__: Union[str, Collection[str]] = ()

    def __init__(*args, __inherit__: Union[str, Collection[str], bool] = (),
                 __exclude__: Union[str, Collection[str]] = (), **kwargs: Callable):
        assert args
        if len(args) > 1:
            raise TypeError('This constructor accepts only keyword arguments.')
        self, = args
        super(Transform, self).__init__(*items_to_container(
            kwargs, type(self), TransformFactory, __inherit__=__inherit__, __exclude__=__exclude__
        ))

    def __repr__(self):
        return f"{self.__class__.__name__}({', '.join(self._methods.fields())})"


class TransformBase(CallableLayer):
    def __init__(self, items: Union[Iterable[Tuple[str, Callable]], Dict[str, Callable]],
                 inherit: Union[str, Collection[str], bool] = (), exclude: Union[str, Collection[str]] = ()):
        super().__init__(*items_to_container(
            items, type(self), TransformFactory, __inherit__=inherit, __exclude__=exclude
        ))


class Mixin(metaclass=APIMeta, __factory=None):
    """
    Base class for all Mixins.
    """

    def __init__(*args, **kwargs):  # noqa
        raise RuntimeError("Mixins can't be directly initialized.")

    __methods__: dict = {}



from typing import Callable, Generator, Any

from .base import Layer
from .chain import connect
from ..cache import MemoryCache
from ..containers import EdgesBag
from ..engine import (
    CustomHash, NodeHashes, NodeHash, Response, Request, Command, StaticGraph, Graph,
    StaticHash, IdentityEdge, FunctionEdge, HashBarrier, Node, CacheEdge, Details, TreeNode
)
from ..exceptions import DependencyError
from ..utils import extract_signature, AntiSet, node_to_dict


class SplitBase(Layer):
    def __init__(self, split: Callable, transform: EdgesBag):
        self._transform = transform
        self._split = split
        self._split_names, _ = extract_signature(split)
        self._part_name = '__part__'
        self._key, self._keys = 'id', 'ids'

    def _make_graph(self, container: EdgesBag, details):
        edges = list(container.edges)
        outputs_mapping = node_to_dict(container.outputs)
        missing = set(self._split_names) - set(outputs_mapping)
        if missing:
            raise DependencyError(
                f'The previous layer is missing the fields {missing}, which are required by the predicate'
            )

        out = Node('__split__pairs__', details)
        edges.append(FunctionEdge(self._split, len(self._split_names)).bind(
            [outputs_mapping[name] for name in self._split_names], out
        ))
        mapping = TreeNode.from_edges(edges)
        return Graph([mapping[container.inputs[0]]], mapping[out])

    def _connect(self, previous: EdgesBag) -> EdgesBag:
        details = Details(type(self))
        previous = previous.freeze(details)

        key = Node(self._key, details)
        keys = Node(self._keys, details)
        key_output = Node(self._key, details)
        key_final = Node(self._key, details)
        edges = [IdentityEdge().bind(key, key_final)]

        previous_outputs = node_to_dict(previous.outputs)
        assert self._key in previous_outputs
        assert self._keys in previous_outputs
        old_keys = previous_outputs[self._keys]

        # build a key -> (left, right) mapping
        mapping = Node('$mapping', details)
        edges.append(SplitMapping(self._make_graph(previous, details)).bind(old_keys, mapping))

        tmp = mapping
        mapping = Node('$mapping', details)
        edges.append(CacheEdge(MemoryCache(None)).bind(tmp, mapping))

        # new keys
        edges.append(FunctionEdge(lambda m: tuple(sorted(m)), 1).bind(mapping, keys))
        # new id to old id
        edges.extend(chain_edges(
            [key, mapping], key_output,
            FunctionEdge(lambda i, m: m[i][0], 2),
            HashBarrier(),
        ))
        # old id part
        part = Node(self._part_name, details)
        edges.append(FunctionEdge(lambda i, m: m[i][1], 2).bind([key, mapping], part))

        return connect(
            # change the key
            EdgesBag(
                [key], [key_output], edges, None, virtual=AntiSet((key_output.name,)),
            ),
            # base layer
            previous,
            # add the __part__
            EdgesBag(
                [], [part], [], None, virtual=AntiSet((part.name,)),
            ),
            # add the transform
            self._transform.freeze(),
            # finally add the new `key` and `keys`
            EdgesBag(
                [], [key_final, keys], [], None, virtual=AntiSet((key_final.name, keys.name, part.name)),
            ),
            freeze=False,
        )


class SplitMapping(StaticGraph, StaticHash):
    def __init__(self, graph: Graph):
        super().__init__(arity=1)
        self.graph = graph
        self._hash = graph.hash()

    def _make_hash(self, inputs: NodeHashes) -> NodeHash:
        return CustomHash('connectome.Split', self._hash, *inputs)

    def evaluate(self) -> Generator[Request, Response, Any]:
        keys = yield Command.ParentValue, 0
        mapping = {}
        for key in keys:
            for new, part in self.graph(key):
                assert new not in mapping, new
                mapping[new] = key, part

        return mapping


def chain_edges(inputs, output, *edges):
    tmp = inputs
    for idx, edge in enumerate(edges, 1):
        inputs = tmp
        tmp = output if idx == len(edges) else Node('$aux')
        yield edge.bind(inputs, tmp)



from collections import deque

from .base import Command


# TODO: replace cache by a thunk tree
def execute(cmd, node, hashes, cache):
    stack, commands = deque([node]), deque([(Command.Return,), (cmd,)])
    push, pop, peek = stack.append, stack.pop, lambda: stack[-1]
    push_command, pop_command = commands.append, commands.pop

    while True:
        cmd, *args = pop_command()

        # return
        if cmd == Command.Return:
            assert len(stack) == 1, len(stack)
            return pop()

        # communicate with edges
        elif cmd == Command.Send:
            node, iterator = args
            value = pop()
            try:
                request = iterator.send(value)

            except StopIteration as e:
                # clear the dependencies
                for n in node.parents:
                    hashes.evict(n)
                    cache.evict(n)
                # return value
                push(e.value)

            else:
                # must continue iteration
                push_command((cmd, node, iterator))
                push_command(request)
                push(node)

        # runs and caches `compute_hash`
        elif cmd == Command.ComputeHash:
            assert not args
            node = pop()
            if node in hashes:
                push(hashes[node])
            else:
                push_command((Command.Store, hashes, node))
                push_command((Command.Send, node, node.edge.compute_hash()))
                push(None)

        # runs and caches `evaluate`
        elif cmd == Command.Evaluate:
            assert not args
            node = pop()
            if node in cache:
                push(cache[node])
            else:
                push_command((Command.Store, cache, node))
                push_command((Command.Send, node, node.edge.evaluate()))
                push(None)

        # requests
        elif cmd == Command.ParentHash:
            idx, = args
            node = pop()

            push_command((Command.Item, 0))
            push_command((Command.ComputeHash,))
            push(node.parents[idx])

        elif cmd == Command.ParentValue:
            idx, = args
            node = pop()

            push_command((Command.Evaluate,))
            push(node.parents[idx])

        elif cmd == Command.CurrentHash:
            assert not args
            push_command((Command.Item, 0))
            push_command((Command.ComputeHash,))

        elif cmd == Command.Payload:
            assert not args
            push_command((Command.Item, 1))
            push_command((Command.ComputeHash,))

        elif cmd == Command.Await:
            node = pop()
            push_command((Command.Tuple, node, len(args), list(args)))

        elif cmd == Command.Call:
            pop()  # pop the node
            func, pos, kw = args
            push(func(*pos, **kw))

        # utils
        elif cmd == Command.Store:
            storage, key = args
            assert key not in storage
            storage[key] = peek()

        elif cmd == Command.Item:
            key, = args
            push(pop()[key])

        elif cmd == Command.Tuple:
            node, n, requests = args
            if not requests:
                push(tuple(pop() for _ in range(n)))
            else:
                request = requests.pop()
                push_command((Command.Tuple, node, n, requests))
                push_command(request)
                push(node)

        else:
            raise RuntimeError('Unknown command', cmd)  # pragma: no cover



from .apply import Apply
from .base import CallableLayer, Chain, Layer, LazyChain, chained
from .cache import CacheLayer, CacheToDisk, CacheToRam
from .check_ids import CheckIds
from .columns import CacheColumns
from .debug import HashDigest
from .filter import Filter
from .group import GroupBy
from .join import Join, JoinMode
from .merge import Merge



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Collection, Dict, Generator, Iterable, NamedTuple, Optional, Sequence, Set, Tuple, Type, Union

from ..exceptions import GraphError
from .node_hash import NodeHash, NodeHashes

__all__ = (
    'Command', 'HashOutput', 'Request', 'Response', 'HashError',
    'Edge', 'Node', 'Nodes', 'NodeSet', 'BoundEdge', 'BoundEdges', 'TreeNode', 'TreeNodes', 'Details',
    'EvalGen', 'HashGen',
)


class Command(Enum):
    ParentHash, CurrentHash, ParentValue, Payload, Await, Call = range(6)
    Send, Store, Item, ComputeHash, Evaluate, Tuple, Return = range(-7, 0)


HashOutput = Tuple[NodeHash, Any]
Request = Tuple  # [RequestType, Any, ...]
Response = Union[NodeHash, Any, Tuple[NodeHash, Any]]
HashGen = Generator[Request, Response, HashOutput]
EvalGen = Generator[Request, Response, Any]


class Edge(ABC):
    def __init__(self, arity: int):
        self.arity = arity

    @abstractmethod
    def compute_hash(self) -> HashGen:
        """ Computes the hash of the output given the input hashes. """

    @abstractmethod
    def evaluate(self) -> EvalGen:
        """ Computes the output value. """

    def hash_graph(self, inputs: NodeHashes) -> NodeHash:
        """ Propagates the graph's hash without any control flow. """
        # TODO: remove this check
        assert len(inputs) == self.arity
        return self._hash_graph(inputs)

    @abstractmethod
    def _hash_graph(self, inputs: NodeHashes) -> NodeHash:
        """ Propagates the graph's hash without any control flow. """

    def bind(self, inputs: Union[Node, Nodes], output: Node) -> BoundEdge:
        if isinstance(inputs, Node):
            inputs = [inputs]
        assert len(inputs) == self.arity, (len(inputs), self.arity)
        return BoundEdge(self, inputs, output)


class Details:
    def __init__(self, layer: Union[str, Type], parent: Optional[Details] = None):
        if isinstance(layer, type):
            layer = layer.__name__
        self.layer = layer
        self.parent = parent

    def update(self, mapping: Dict[Details, Details], parent: Union[Details, None]):
        """ Update the whole tree based on a mapping """
        assert isinstance(parent, Details) or parent is None, parent
        if self.parent is not None:
            if self.parent in mapping:
                parent = mapping[self.parent]
            else:
                parent = mapping[self.parent] = self.parent.update(mapping, parent)

        return Details(self.layer, parent)

    def __str__(self):
        result = f'{self.layer} ({hex(id(self))})'
        if self.parent is not None:
            result += ' -> ' + str(self.parent)
        return result


class TreeNode:
    __slots__ = 'name', '_edge', 'details'

    def __init__(self, name: str, edge: Optional[Tuple[Edge, Sequence[TreeNode]]],
                 details: Union[Details, None] = None):
        self.name, self._edge, self.details = name, edge, details

    @property
    def is_leaf(self):
        return self._edge is None

    @property
    def edge(self):
        return self._edge[0]

    @property
    def parents(self):
        return self._edge[1]

    @classmethod
    def from_edges(cls, edges: Iterable[BoundEdge]) -> Dict[Node, TreeNode]:
        def update(node: Node):
            if node in mapping:
                return mapping[node]

            bridge = bridges.get(node)
            if bridge is not None:
                bridge = bridge.edge, tuple(map(update, bridge.inputs))
            mapping[node] = new = cls(node.name, bridge, node.details)
            return new

        nodes = set()
        bridges = {}
        # each edge is represented by its output
        for edge in edges:
            if edge.output in bridges:
                raise GraphError(f'The node "{edge.output.name}" has multiple incoming edges')

            bridges[edge.output] = edge
            nodes.add(edge.output)
            nodes.update(edge.inputs)

        mapping = {}
        for n in nodes:
            update(n)
        return mapping

    @staticmethod
    def to_edges(nodes: Iterable[TreeNode]) -> Sequence[BoundEdge]:
        def reverse(node) -> Node:
            if node not in _reversed:
                _reversed[node] = Node(node.name, node.details)
            return _reversed[node]

        def visit(node: TreeNode):
            if node in visited or node.is_leaf:
                return
            visited.add(node)
            for parent in node.parents:
                visit(parent)

            result.append(BoundEdge(node.edge, list(map(reverse, node.parents)), reverse(node)))

        result = []
        visited = set()
        _reversed = {}
        for n in nodes:
            visit(n)

        return tuple(result)

    def __repr__(self):
        return f'<TreeNode: {self.name}>'


class Node:
    __slots__ = 'name', 'details'

    def __init__(self, name: str, details: Union[Details, None] = None):
        assert isinstance(details, Details) or details is None, details
        self.name = name
        self.details = details

    def clone(self):
        return type(self)(self.name, self.details)

    def __repr__(self):
        return f'<Node: {self.name}>'


class BoundEdge(NamedTuple):
    edge: Edge
    inputs: Nodes
    output: Node


TreeNodes = Collection[TreeNode]
Nodes = Collection[Node]
NodeSet = Set[Node]
BoundEdges = Collection[BoundEdge]
Edges = Collection[Edge]


class HashError(RuntimeError):
    pass



from typing import Union, Collection, Callable, Iterable

from ..containers import EdgesBag
from ..engine import IdentityEdge
from ..interface.factory import TransformFactory
from ..interface.metaclasses import APIMeta
from ..layers.split import SplitBase


class SplitFactory(TransformFactory):
    _part_name = '__part__'
    _split_name = '__split__'
    layer_cls = SplitBase

    def __init__(self, layer: str, scope):
        self._split: Callable = None
        super().__init__(layer, scope)

    def _prepare_layer_arguments(self, container: EdgesBag, properties: Iterable[str]):
        assert not properties, properties
        return self._split, container

    def _before_collect(self):
        super()._before_collect()
        self.edges.append(IdentityEdge().bind(self.inputs[self._part_name], self.parameters[self._part_name]))
        self.magic_dispatch[self._split_name] = self._handle_split

    def _handle_split(self, value):
        assert self._split is None, self._split
        self._split = value

    def _after_collect(self):
        super()._after_collect()
        assert self._split is not None
        assert not self.special_methods, self.special_methods


# TODO: Examples
class Split(SplitBase, metaclass=APIMeta, __factory=SplitFactory):
    """
    Split a dataset entries into several parts.

    This layer requires a `__split__` magic method, which takes an entry id, and returns a list of parts -
    (part_id, part_context) pairs, "part_id" will become the part's id, and "part_context" is accessible in other
    methods as part-specific useful info.
    """

    __inherit__: Union[str, Collection[str], bool] = ()
    __exclude__: Union[str, Collection[str]] = ()

    def __init__(self, *args, **kwargs):
        raise NotImplementedError

    def __split__(*args, **kwargs):
        raise NotImplementedError
